% Computes the local centers of mass (CMs) along the second dimension of a
% multi-dimensional array.
%
% [CM, dist] = findCMs(dI, alpha, dim)
%
% dI:     Magnitude of the derivative of an image (multi-dimensional array)
%         along its second dimension, raised to a power. For instance, for
%         an image 'I', 'dI' could be diff(I,1,dim).^2 (see how findCMs is
%         used in segCM.m).
% alpha:  Weighting of the edges.
% dim:    The dimension through which the CMs are computed (default: 1).
%
% CM:     Local CMs computed along the second dimension of 'dI'.
% dist:   The distance from each voxel to its computed local CM.
%
% See also:  EXAMPLE, segCM.

% Codes by Iman Aganj.
%
% I. Aganj, M. G. Harisinghani, R. Weissleder, and B. Fischl,
% "Unsupervised medical image segmentation based on the local center of mass",
% Scientific Reports, vol. 8, Article no. 13012, 2018.
% www.nature.com/articles/s41598-018-31333-5

function [CM, dist] = findCMs(dI, alpha, dim)

if ~exist('dim', 'var')
    dim = 1;
end
oldVer = verLessThan('matlab', '9.1'); % Matlab releases earlier than R2016b.
reshapeArg = ones(1,ndims(dI));
reshapeArg(dim) = size(dI, dim);
x0 = reshape(1:size(dI, dim), reshapeArg);
if isa(dI, 'gpuArray')
    x0 = gpuArray(x0);
end
mX = mean(x0);
x = x0 - mX;
D = alpha * cumsum(dI, dim);
clear dI
if oldVer
    D = bsxfun(@minus, D, mean(D, dim));
else
    D = D - mean(D, dim);
end
epD = exp(D);
enD = exp(-D);
clear D
if oldVer
    CM = (enD.*cumsum(bsxfun(@times, x, epD), dim) + epD.*cumsum(bsxfun(@times, x, enD), dim, 'reverse')) ./ (enD.*cumsum(epD, dim) + epD.*cumsum(enD, dim, 'reverse')) + mX;
else
    CM = (enD.*cumsum(x.*epD, dim) + epD.*cumsum(x.*enD, dim, 'reverse')) ./ (enD.*cumsum(epD, dim) + epD.*cumsum(enD, dim, 'reverse')) + mX;
end
clear epD enD
CM(isinf(CM)) = nan;

if nargout > 1
    if oldVer
        dist = abs(bsxfun(@minus, CM, x0));
    else
        dist = abs(CM - x0);
    end
end
