% Segments an image into labels using a center-of-mass based approach.
%
% [L, distCM] = segCM(I, param, L)
%  
% I:          2D or 3D input image.
% param:      Optional structure containing the following parameters.
%             A default value is set to any non-existing field.
%    param.alpha:        Weighting of the edges (default: 2000, for an
%                        image with intensities between 0 and 1).
%    param.maxIter:      Maximum number of iterations, in case the
%                        algorithm has not converged prior to that
%                        (default: 1000 for 2D and 500 for 3D; use
%                        smaller numbers for faster results).
%    param.randIterNum:  Number of iterations in the first (random) phase
%                        of segmentation (default: 100 for 2D, 50 for 3D).
%    param.power:        Norm of the gradient (default: 2).
%    param.angleStep:    Angular resolution of the orientations in degrees
%                        (default: 1 for 2D, 10 for 3D).
%    param.indexGPU:     If provided, the GPU device specified by this
%                        index (e.g., 1) will be used.
%    param.saveFile:     If provided, final and intermediate results will
%                        be saved in the "Results" dirctory, with
%                        param.saveFile as the filename prefix.
%    param.saveIter:     if param.saveFile has been provided, intermediate
%                        results will be saved every param.saveIter
%                        iterations (default: 200 for 2D and 50 for 3D).
%    param.showFigure:   If 'true' for 2D segmentation, the results will
%                        be visualized on-the-fly (default: false).
%    param.dim:          Number of dimensions of the image (default: number
%                        of dimensions of 'I'). If param.dim is specified
%                        as one less than the number of dimensions of 'I',
%                        the image will be assumed to be multichannel. For
%                        instance, if 'I' is a 2D RGB image and 100x100x3,
%                        then the user must specify param.dim=2 to indicate
%                        that 'I' is not a 3D image.
%    param.verbose:      If 'true' (default), iteration numbers will be
%                        shown.
% L (input):  Initial label image (optional).
%             Choosing the L='intermediate' will cause the algorithm to
%             continue the segmentation from the saved intermediate results.
%
% L (output): Resulting label image.
% distCM:     The mean distance of the pixels from their CMs.
%
% See also:  EXAMPLE, findCMs.

% Codes by Iman Aganj.
%
% Reference:
% I. Aganj, M. G. Harisinghani, R. Weissleder, and B. Fischl,
% "Unsupervised medical image segmentation based on the local center of mass",
% Scientific Reports, vol. 8, Article no. 13012, 2018.
% www.nature.com/articles/s41598-018-31333-5

function [L, distCM] = segCM(I, param, L)

if ischar(I)
    load(I, 'I')
end
if ~exist('param', 'var')
    param = [];
end
existL = exist('L', 'var');
if ~isfield(param, 'dim')
    if ndims(I)==4
        param.dim = 3;
    else
        param.dim = ndims(I);
    end
end
D = param.dim;
if ~ismember(D, 1:3)
    error('The number of image dimensions can only be 2 or 3!')
end
if ~isfield(param, 'power')
    param.power = 2;
end
if ~isfield(param, 'alpha')
    param.alpha = 2000 / (max(I(:))-min(I(:)))^param.power;
end
if ~isfield(param, 'maxIter')
    if D==2
        param.maxIter = 1000;
    else
        param.maxIter = 500;
    end
end
if ~isfield(param, 'randIterNum')
    if D==2
        param.randIterNum = 100;
    else
        param.randIterNum = 50;
    end
end
if ~isfield(param, 'angleStep')
    if D==2
        param.angleStep = 1;
    else
        param.angleStep = 10;
    end
end
if ~isfield(param, 'indexGPU')
    param.indexGPU = false;
end
if ~isfield(param, 'saveFile')
    param.saveFile = [];
end
if isempty(param.saveFile)
    saveFileAll = [];
else
    saveFileAll = [param.saveFile '_alpha' num2str(param.alpha) '_rand' num2str(param.randIterNum)];
    if exist('Results', 'file')~=7
        mkdir('Results')
        disp('The ''Results'' folder was created.')
    end
end
if ~isfield(param, 'saveIter')
    if D==2
        param.saveIter = 200;
    else
        param.saveIter = 50;
    end
end
if ~isfield(param, 'showFigure')
    param.showFigure = false;
end
if param.showFigure && D>2
    warning('showFigure can only be used for 2D images!')
    param.showFigure = false;
end
if ~isfield(param, 'verbose')
    param.verbose = true;
end
if ~isfield(param, 'crop')
    param.crop = true; % Keep 'true' for faster performance.
end

if param.verbose
    disp('Parameters:')
    disp(param)
end

s0 = size(I);
s0 = s0(1:D);
nChannels = size(I, D+1);
if D==3 && s0(3)==3 && nChannels==1
    disp(['Assuming ''I'' to be a 3D image of size ' num2str(s0(1)) 'x' num2str(s0(2)) 'x3. If it is a 2D RGB image, set param.dim=2.'])
end
mSize = max(s0);
I = double(I);
outsideValue = I;
for d = 1:D
    outsideValue = max(outsideValue, [], d);
end
outsideValue = 2 * outsideValue * param.crop;
cI = bsxfun(@times, ones(mSize*ones(1,D)), outsideValue);
cropX = arrayfun(@(siz) round((mSize-siz)/2) + (1:siz), s0', 'UniformOutput', false);
cI(cropX{:},:) = I;
if param.showFigure
    subplot(1,2,1)
    if nChannels==2
        if param.crop
            image(cat(3, I, zeros(size(I,1),size(I,2))))
        else
            image(cat(3, cI, zeros(size(cI,1),size(cI,2))))
        end
    elseif nChannels==3
        if param.crop
            image(I)
        else
            image(cI)
        end
    else
        if param.crop
            imagesc(mean(I,  3))
        else
            imagesc(mean(cI, 3))
        end
        colormap(gca, gray)
    end
    axis equal tight
    drawnow
end
clear I

if D==2
    angles = 0:param.angleStep:179;
else
    for theta = 0:param.angleStep:89
        if theta == 0
            angles = [0; 0];
        else
            phi = 0:(param.angleStep/sin(theta*pi/180)):359;
            angles = [angles, [theta*ones(1,length(phi)); phi]];
        end
    end
    rotAxis = [-sin(angles(2,:)*pi/180); zeros(1,size(angles,2)); cos(angles(2,:)*pi/180)]';
end

sBox = single(size(cI));
sBox = sBox(1:D);
if param.crop
    s = s0;
else
    s = sBox;
end
sA = size(angles,2);
padZeros = zeros([1 double(sBox(2:end))]);
X = cell(D,1);
cellOne2S = arrayfun(@(siz) 1:siz, sBox, 'UniformOutput', false);
[X{:}] = ndgrid(cellOne2S{:});
X = cellfun(@uint16, X, 'UniformOutput', false);
if param.indexGPU
    gpu = gpuDevice(param.indexGPU);
    if param.verbose
        disp(['Using the "' gpu.Name '" GPU...'])
    end
    if D==2
        X = cellfun(@gpuArray, X, 'UniformOutput', false);
    end
end
aX = zeros([sA s D], 'uint16');
aD = zeros([sA s]  , 'uint16');

if license('test', 'Distrib_Computing_Toolbox')
    pool = gcp('nocreate'); % If a parpool is available, it will be used.
else
    pool = [];
end
if isempty(pool)
    nCores = 0;
else
    nCores = pool.NumWorkers;
end
cI = bsxfun(@minus, cI, outsideValue);
if param.verbose
    disp('Creating the CM images...')
end
tic
if D==2
    parfor (a = 1:sA, nCores) % To not require the Parallel Computing Toolbox, change to:   for a = 1:sA
        if param.verbose
            disp(['Direction # ' num2str(a) ' / ' num2str(sA)])
        end
        Irot_D = bsxfun(@plus, imrotate(cI, angles(a), 'bilinear', 'crop'), outsideValue);
        if param.indexGPU
            Irot_D = gpuArray(Irot_D);
        end
        if nChannels==1
            Irot_D = [abs(diff(Irot_D,1,1)).^param.power; padZeros];
        else
            Irot_D = [mean(abs(diff(Irot_D,1,1)).^param.power, D+1); padZeros];
        end
        [C, Irot_D] = findCMs(Irot_D, param.alpha); % Irot_D is now the distance from the CM.
        C = round(single(C));
        C(C<1 | C>mSize) = nan;
        Irot_D = gather(uint16(round(imrotate(single(Irot_D), -angles(a), 'bilinear', 'crop'))));
        if param.crop
            aD(a,:,:) = Irot_D(cropX{:});
        else
            aD(a,:,:) = Irot_D;
        end
        C = sub2ind(sBox, C(:), X{2}(:)); % C is now the CM indices.
        nanC = isnan(C);
        C(nanC) = 1;
        for d = 1:D
            Xrot = imrotate(X{d}, angles(a), 'nearest', 'crop');
            Xrot = Xrot(C);
            Xrot(nanC | isnan(Xrot)) = 0;
            Xrot = gather(imrotate(reshape(Xrot, sBox), -angles(a), 'nearest', 'crop'));
            if param.crop
                aX(a,:,:,d) = Xrot(cropX{:}) - cropX{d}(1) + 1;
            else
                aX(a,:,:,d) = Xrot;
            end
        end
    end
else
    parfor (a = 1:sA, nCores) % To not require the Parallel Computing Toolbox, change to:   for a = 1:sA
        if param.verbose
            disp(['Direction # ' num2str(a) ' / ' num2str(sA)])
        end
        if nChannels==1
            Irot_D = imrotate3(cI, angles(1,a), rotAxis(a,:), 'linear', 'crop') + outsideValue;
            Irot_D = [abs(diff(Irot_D,1,1)).^param.power; padZeros];
        else
            Irot_D = zeros([sBox nChannels], 'like', cI);
            for i = 1:nChannels
                Irot_D(:,:,:,i) = imrotate3(cI(:,:,:,i), angles(1,a), rotAxis(a,:), 'linear', 'crop') + outsideValue(i);
            end
            Irot_D = [mean(abs(diff(Irot_D,1,1)).^param.power, D+1); padZeros];
        end
        if param.indexGPU
            Irot_D = gpuArray(Irot_D);
        end
        [C, Irot_D] = findCMs(Irot_D, param.alpha);  % Irot_D is now the distance from the CM.
        C = round(single(C));
        C(C<1 | C>mSize) = nan;
        Irot_D(isnan(Irot_D)) = -1e10;
        Irot_D = uint16(round(imrotate3(gather(Irot_D), -angles(1,a), rotAxis(a,:), 'linear', 'crop')));
        if param.crop
            aD(a,:,:,:) = Irot_D(cropX{:});
        else
            aD(a,:,:,:) = Irot_D;
        end
        C = sub2ind(sBox, C(:), X{2}(:), X{3}(:)); % C is now the CM indices.
        nanC = isnan(C);
        C(nanC) = 1;
        for d = 1:D
            Xrot = imrotate3(X{d}, angles(1,a), rotAxis(a,:), 'nearest', 'crop');
            Xrot = Xrot(C);
            Xrot(nanC | isnan(Xrot)) = 0;
            Xrot = imrotate3(reshape(Xrot, sBox), -angles(1,a), rotAxis(a,:), 'nearest', 'crop');
            if param.crop
                aX(a,:,:,:,d) = Xrot(cropX{:}) - cropX{d}(1) + 1;
            else
                aX(a,:,:,:,d) = Xrot;
            end
        end
    end
end
t = toc;
if param.verbose
    disp(['CM images were created for ' num2str(sA) ' directions in ' num2str(t) ' seconds.'])
end
clear I Irot_D Xrot C nanC

mask = ~any(isnan(aX), D+2) & all(aX>=1, D+2) & all(bsxfun(@le, aX, permute(s, [3:(2+D), 1 2])), D+2);
aD = single(aD);
mask = mask & bsxfun(@times, rand([sA s], 'single'), max(aD, [], 1).^(D-1)) <= aD.^(D-1);
if nargout > 1
    %aD(~mask) = nan;
    if ~existL
        L = [];
    end
    distCM = squeeze(nanmean(aD,1));
    return
end
delete(pool)
clear aD

showLocalCMs = false;
if showLocalCMs
    if param.crop
        I = I(cropX{:});
    end
    if D==2
        J = cat(3, I, zeros([s 2]));
        while true
            image(J), axis equal tight; drawnow
            [y,x] = ginput(1);
            [x,y] = deal(round(x), round(y));
            J(:,:,3) = 0; J(x,y,3) = 1;
            J0 = zeros(s);
            J0(sub2ind(s, aX(mask(:,x,y),x,y,1), aX(mask(:,x,y),x,y,2))) = 1;
            J(:,:,2) = J0;
        end
    else
        while true
            r = input('[x,y,z] ? ');
            [x,y,z] = deal(round(r(1)), round(r(2)), round(r(3)));
            J = false(s);
            J(sub2ind(s, aX{1}(mask(:,x,y,z),x,y,z), aX{2}(mask(:,x,y,z),x,y,z), aX{3}(mask(:,x,y,z),x,y,z))) = true;
            showvol(I+J, true) % 'implay' can also be used.
        end
    end
end
clear cI

aX = num2cell(aX, 1:(D+1));
aX = cellfun(@(c) c(mask(:)), aX, 'UniformOutput', false);
indL = uint32(sub2ind(s, aX{:}));
clear aX
nMask = single(squeeze(sum(mask,1)));
indM = nMask(:)>0;
if param.crop
    X = cellfun(@(c, cr) c(cropX{:}) - cr(1) + 1, X, cropX, 'UniformOutput', false);
    cropX = [];
end
X = cellfun(@(c) c(indM), X, 'UniformOutput', false);
nMask = nMask(indM); sNMask = length(nMask);
mask = false([sA s]);
mask(sub2ind([sA s], nMask, X{:})) = true;
mask = cumsum(mask, 1, 'reverse')>0;

nIter = 1;
if existL
    if strcmpi(L, 'intermediate')
        if isempty(saveFileAll)
            error('param.saveFile needs to be specified to retrieve intermediate results!')
        else
            fileL0 = fullfile('Results', [saveFileAll '_intermediate.mat']);
            if exist(fileL0, 'file')
                load(fileL0, 'L', 'nIter')
            else
                disp(['The intermediate file (' fileL0 ') wasn''t found!'])
                L = nan(s, 'single');
            end
        end
    end
    L = single(L);
else
    L = nan(s, 'single');
end
r = setdiff(1:prod(s), unique(L(:)));
L(isnan(L(:))) = r(randperm(nnz(isnan(L(:)))));
clear r
if param.showFigure
    cMap = hsv(min(8192, max(L(:))));
    if existL
        cMap = cMap(randperm(size(cMap,1)),:);
    end
    subplot(1,2,2), colormap(gca, cMap)
end
if param.indexGPU
    s = gpuArray(s);
    sA = gpuArray(sA);
    mask = gpuArray(mask);
    nMask = gpuArray(nMask);
    indL = gpuArray(indL);
    indM = gpuArray(indM);
    L = gpuArray(L);
end

if param.verbose
    disp('Random iterations...')
end
maskSize = find(~any(any(any(mask,2),3),4), 1, 'first');
if isempty(maskSize)
    sT = [sA s];
else
    sT = [maskSize-1, s];
    mask = mask(1:sT(1),:,:,:);
end
keepWhile = true;
while keepWhile && nIter<=param.maxIter
    verbPhrase = ['Iteration # ' num2str(nIter) ' out of ' num2str(param.maxIter) '.'];
    if param.verbose
        disp(verbPhrase)
    end
    L0 = L;
    aL = nan(sT, 'like', L);
    aL(mask(:)) = L(indL);
    if nIter > param.randIterNum
        L = squeeze(mode(aL, 1));
    else
        L(indM) = aL(sub2ind(sT, round(rand(sNMask, 1, 'like', s) .* (nMask-1)) + 1, X{:}));
        if nIter == param.randIterNum
            if param.verbose
                disp('Random phase finished.')
            end
        end
    end
    nanL = isnan(L);
    L(nanL) = L0(nanL);
    if isequaln(L,L0)
        keepWhile = false;
    end
    [~, ~, L(:)] = unique(L(:));
    if param.showFigure
        imagesc(L)
        axis equal tight
        colormap(gca, cMap)
        set(gcf, 'Name', verbPhrase)
        drawnow
    end
    if ~isempty(param.saveFile) && mod(nIter,param.saveIter) == 0
        saveResults(L, nIter, param, cropX, [saveFileAll '_intermediate']);
    end
    nIter = nIter + 1;
end
if param.verbose
    disp(['Segmentation took ' num2str(toc) ' seconds.'])
end
if param.showFigure
    edgeL = L;
    edgeL([diff(L,1,1); zeros(1,size(L,2))] | [diff(L,1,2) zeros(size(L,1),1)]) = 0;
    imagesc(edgeL), colormap(gca, [0 0 0; cMap]), axis equal tight
end
L = saveResults(L, nIter, param, cropX, saveFileAll);
if ~isempty(param.saveFile)
    delete(fullfile('Results', [saveFileAll '_intermediate.mat']));
end

end


function L = saveResults(L, nIter, param, cropX, saveFileAll)

if ~isempty(cropX)
    L = L(cropX{:});
end
[~, ~, L(:)] = unique(L(:));
L = gather(L);
t = toc;
if ~isempty(saveFileAll)
    filename = fullfile('Results', [saveFileAll '.mat']);
    save(filename, 'L', 'nIter', 'param', 't')
    disp(['Results saved to: ' filename])
end
end
